"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from typing import Tuple, Callable

import numpy as np
import jax
import jax.numpy as jnp
from flax import linen as nn
import chex

import minimax.envs as envs
from .rnn import CustomOptimizedLSTMCell


def calc_gain(kind):
	if kind == 'linear':
		return 1.0
	elif kind == 'conv':
		return 1.0
	elif kind == 'sigmoid':
		return 1.0
	elif kind == 'tanh':
		return np.sqrt(2)
	elif kind == 'relu':
		return np.sqrt(2)
	elif kind == 'leaky_relu':
		return np.sqrt(2/(1+0.01))
	elif kind == 'selu':
		return 0.75
	elif kind == 'gelu':
		return 0.75 
	elif kind == 'crelu':
		return np.sqrt(2)


def crelu(x):
	return jnp.concatenate((nn.relu(x), nn.relu(-x)), axis=-1)


def get_activation(name):
	if name == 'crelu':
		return crelu
	else:
		return getattr(nn, name)


def default_bias_init(scale=1.0):
	return nn.initializers.zeros


def init_orth(scale=1.0):
	return nn.initializers.orthogonal(scale)


def make_fc_layers(
	name=None, 
	n_layers=1, 
	hidden_dim=32,
	activation=None, 
	kernel_init=None,
	use_layernorm=False):
	if kernel_init is None:
		kernel_init = init_orth(
			scale=calc_gain('linear')
		)

	layers = []
	for i in range(n_layers):
		layer_name = None
		if name:
			layer_name = f'{name}_{i+1}'

		layers.append(
			nn.Dense(
				hidden_dim, 
				kernel_init=kernel_init,
				name=layer_name,
			)
		)

		if use_layernorm:
			layers.append(nn.LayerNorm())
		
		if activation is not None:
			layers.append(activation)

	return nn.Sequential(layers)


def make_rnn(
	arch='lstm', 
	kernel_init=init_orth(), 
	recurrent_kernel_init=init_orth(),
	name=None):
	if arch == 'lstm':
		rnn = CustomOptimizedLSTMCell(
			kernel_init=init_orth(),
			recurrent_kernel_init=init_orth(),
			name=name
		)
	elif arch == 'gru':
		rnn = nn.GRUCell(
			kernel_init=init_orth(),
			recurrent_kernel_init=init_orth(),
			name=name
		)
	else:
		rnn = None

	return rnn


class RecurrentModuleBase(nn.Module):
	def initialize_carry(
			self, 
			rng: chex.PRNGKey, 
			batch_dims: Tuple[int] = ()) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
		"""Initialize hidden state of LSTM."""
		if self.recurrent_arch == 'lstm':
			return nn.OptimizedLSTMCell.initialize_carry(
				rng, batch_dims, self.recurrent_hidden_dim
			)
		elif self.recurrent_arch == 'gru':
			return nn.GRUCell.initialize_carry(
				rng, batch_dims, self.recurrent_hidden_dim
			)
		else:
			raise ValueError('Model is not recurrent.')

	@property
	def is_recurrent(self):
		return self.recurrent_arch is not None


class ScannedRNN(nn.Module):
	"""
	Scanned RNN.
	Inputs:
		carry: time-major input hidden states, LxBxH and optional
		resets: Reset flags of shape LxB, where 1 indicates reset (equivalent to done==True).
	"""
	recurrent_arch: str = 'lstm'
	recurrent_hidden_dim: int = 256
	kernel_init: Callable = init_orth()
	recurrent_kernel_init: Callable = init_orth()

	@partial(
		nn.scan,
		variable_broadcast="params",
		in_axes=0,
		out_axes=0,
		split_rngs={"params": False},
	)
	@nn.compact
	def __call__(self, carry, step):
		x, reset = step
		rnn_state = carry

		rnn_state = jax.tree_map(
			lambda x,y: jax.vmap(jax.lax.select)(reset, x, y), 
			ScannedRNN.initialize_carry(
				jax.random.PRNGKey(0), (x.shape[0],), self.recurrent_hidden_dim, self.recurrent_arch),	
			rnn_state
		)

		rnn_kwargs = dict(
			features=self.recurrent_hidden_dim,
			kernel_init=self.kernel_init,
			recurrent_kernel_init=self.recurrent_kernel_init,
		)
		if self.recurrent_arch == 'lstm':
			rnn_cell = nn.OptimizedLSTMCell(**rnn_kwargs) # defaults to orth init
		elif self.recurrent_arch == 'gru':
			rnn_cell = nn.GRUCell(**rnn_kwargs)
		else:
			raise ValueError(f'Unsupported recurrent_arch={self.recurrent_arch}')

		new_rnn_state, y = rnn_cell(rnn_state, x)
		return new_rnn_state, y

	@staticmethod
	def initialize_carry(rng, batch_dims, recurrent_hidden_dim, recurrent_arch):
		init_args = (rng, (*batch_dims, recurrent_hidden_dim))
		if recurrent_arch == 'lstm':
			return nn.OptimizedLSTMCell(recurrent_hidden_dim, parent=None).initialize_carry(*init_args) # defaults to orth init
		elif recurrent_arch == 'gru':
			return nn.GRUCell(recurrent_hidden_dim, parent=None).initialize_carry(*init_args)
		else:
			raise ValueError(f'Unsupported recurrent_arch={recurrent_arch}')


class ValueHead(nn.Module):
	n_hidden_layers: int = 1
	hidden_dim: int = 256
	activation: Callable = nn.tanh
	kernel_init: Callable = init_orth(calc_gain('tanh'))
	last_layer_kernel_init: Callable = init_orth(calc_gain('linear'))
	use_layernorm: bool = False

	@nn.compact
	def __call__(self, x):
		return nn.Sequential([
			make_fc_layers(
				n_layers=self.n_hidden_layers, 
				hidden_dim=self.hidden_dim,
				activation=self.activation,
				kernel_init=self.kernel_init,
				use_layernorm=self.use_layernorm
			),
			nn.Dense(
				1, 
				kernel_init=self.last_layer_kernel_init,
				name='fc_value_final'
			)
		])(x)


class EnsembleValueHead(nn.Module):
	n: int = 2

	n_hidden_layers: int = 1
	hidden_dim: int = 256
	activation: Callable = nn.tanh
	kernel_init: Callable = init_orth(calc_gain('tanh'))
	last_layer_kernel_init: Callable = init_orth(calc_gain('linear'))

	@nn.compact
	def __call__(self, x):
		"""
		Assumes x is batch
		"""
		VmapValue = nn.vmap(
			ValueHead,
			variable_axes={"params": 0},
			split_rngs={"params": True},
			in_axes=None,
			out_axes=1,
			axis_size=self.n,
		)
		vs = VmapValue(
			n_hidden_layers=self.n_hidden_layers,
			hidden_dim=self.hidden_dim,
			activation=self.activation,
			kernel_init=self.kernel_init,
			last_layer_kernel_init=self.last_layer_kernel_init
		)(x)

		return vs


def clean_init_kwargs_prefix(prefix):
	def class_decorator(cls):
		old_init = cls.__init__
		def new_init(self, *args, **kwargs):
			kwargs = {
				k.removeprefix(prefix):v for k,v in kwargs.items()
			}
			old_init(self, *args, **kwargs)

		cls.__init__ = new_init
		return cls
		
	return class_decorator



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from .xp_runner import ExperimentRunner
from .eval_runner import EvalRunner
from .dr_runner import DRRunner
from .plr_runner import PLRRunner
from .paired_runner import PAIREDRunner


__all__ = [
	ExperimentRunner,
	EvalRunner,
	DRRunner,
	PLRRunner,
	PAIREDRunner
]


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import chex
from typing import Tuple, Union, Optional

from .env_wrapper import EnvWrapper
from minimax.envs.environment import EnvState


class MonitorReturnWrapper(EnvWrapper):
	"""
	Tracks episodic returns and, optionally, environment metrics.
	"""
	def reset_extra(self):
		if self._wrap_level > 1:
			extra = self._env.reset_extra()
		else:
			extra = {}

		extra.update({
			'ep_return': 0.,
		})

		return extra

	def get_monitored_metrics(self):
		return super().get_monitored_metrics() + ('return',)

	def step(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
		reset_state: Optional[chex.ArrayTree] = None,
		extra: dict = None,
	) -> Tuple[chex.Array, EnvState, float, bool]:
		step_kwargs = dict(
			reset_state=reset_state
		)
		if self._wrap_level > 1:
			step_kwargs.update(dict(
				extra=extra
			))

		step = self._env.step(
				key, 
				state, 
				action, 
				**step_kwargs)

		if len(step) == 5:
			obs, state, reward, done, info = step
		else:
			obs, state, reward, done, info, extra = step

		# Track returns
		extra['ep_return'] += reward
		info['return'] = done*extra['ep_return']
		extra['ep_return'] *= (1-done)

		return obs, state, reward, done, info, extra



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import os
import json
import re
import fnmatch
import sys
from collections import defaultdict

import numpy as np
import pandas as pd
import scipy.stats as spstats
import jax
import jax.numpy as jnp
from tqdm import tqdm

from minimax.util.parsnip import Parsnip
from minimax.util.args import str2bool
from minimax.util.checkpoint import load_pkl_object, load_config
from minimax.util.dotdict import DefaultDotDict
from minimax.util.loggers import HumanOutputFormat
from minimax.util.rl import AgentPop
from minimax.runners import EvalRunner
import minimax.models as models
import minimax.agents as agents


parser = Parsnip()

# ==== Define top-level arguments
parser.add_argument(
	'--seed', 
	type=int, 
	default=1, 
	help='Random seed.')
parser.add_argument(
	'--log_dir', 
	type=str,
	default='~/logs/minimax', 
	help='Log directory containing experiment dirs.')
parser.add_argument(
	'--xpid', 
	type=str, 
	default='latest', 
	help='Experiment ID dir name for model.')
parser.add_argument(
	'--xpid_prefix', 
	type=str, 
	default=None, 
	help='Experiment ID dir name for model.')
parser.add_argument(
	'--checkpoint_name', 
	type=str, 
	default='checkpoint', 
	help='Name of checkpoint .pkl.')
parser.add_argument(
	'--env_names', 
	type=str, 
	help='csv of evaluation environments.')
parser.add_argument(
	'--n_episodes', 
	type=int, 
	default=1, 
	help='Number of evaluation episodes.')
parser.add_argument(
	'--agent_idxs', 
	type=str, 
	default='*', 
	help="Indices of agents to evaluate. '*' indicates all.")
parser.add_argument(
    '--render_mode',
    type=str, 
    nargs='?', 
    const=True, 
    default=None,
    help='Visualize episodes.')
parser.add_argument(
    '--results_path',
    type=str,   
    default='results/',
    help='Results dir.')
parser.add_argument(
    '--results_fname',
    type=str,   
    default=None,
    help='Results filename (without .csv).')


if __name__ == '__main__':
	"""
	Usage: 
		python -m eval \
		--xpid= \
		--env_names="Maze-SixteenRooms" \
		--n_episodes=100 \
		--agent_idxs=0
	"""
	args = parser.parse_args()

	log_dir_path = os.path.expandvars(os.path.expanduser(args.log_dir))

	xpids = []
	if args.xpid_prefix is not None:
		# Get all matching xpid directories
		all_xpids = fnmatch.filter(os.listdir(log_dir_path), f"{args.xpid_prefix}*")
		filter_re = re.compile('.*_[0-9]*$')
		xpids = [x for x in all_xpids if filter_re.match(x)]
	else:
		xpids = [args.xpid]

	pbar = tqdm(total=len(xpids))

	all_eval_stats = defaultdict(list)
	for xpid in xpids:
		xpid_dir_path = os.path.join(log_dir_path, xpid)
		checkpoint_path = os.path.join(xpid_dir_path, f'{args.checkpoint_name}.pkl')
		meta_path = os.path.join(xpid_dir_path, f'meta.json')

		# Load checkpoint info
		if not os.path.exists(meta_path):
			print(f'Configuration at {meta_path} does not exist. Skipping...')
			continue

		if not os.path.exists(checkpoint_path):
			print(f'Checkpoint path {checkpoint_path} does not exist. Skipping...')
			continue
		
		xp_args = load_config(meta_path)

		agent_idxs = args.agent_idxs
		if agent_idxs == '*':
			agent_idxs = np.arange(xp_args.train_runner_args.n_students)
		else:
			agent_idxs = \
				np.array([int(x) for x in agent_idxs.split(',')])
			assert np.max(agent_idxs) <= xp_args.train_runner_args.n_students, \
				'Agent index is out of bounds.'

		runner_state = load_pkl_object(checkpoint_path)
		params = runner_state[1]['params']
		params = jax.tree_util.tree_map(
			lambda x: jnp.take(x, indices=agent_idxs, axis=0),
			params
		)

		with jax.disable_jit(args.render_mode is not None):
			student_model = models.make(
				env_name=xp_args.env_name,
				model_name=xp_args.student_model_name,
				**xp_args.student_model_args
			)

			pop = AgentPop(
				agent=agents.PPOAgent(model=student_model),
				n_agents=len(agent_idxs)
			)

			# We force EvalRunner to select all params, since we've already
			# extracted the relevant agent indices.
			runner = EvalRunner(
				pop=pop,
				env_names=args.env_names,
				env_kwargs=xp_args.eval_env_args,
				n_episodes=args.n_episodes,
				render_mode=args.render_mode,
				agent_idxs='*'
			)

			rng = jax.random.PRNGKey(args.seed)
			_eval_stats = runner.run(rng, params)

			eval_stats = {}
			for k,v in _eval_stats.items():
				prefix_match = re.match(r'^eval/(a[0-9]+):.*', k)
				if prefix_match is not None:
					prefix = prefix_match.groups()[0]
					_idx = int(prefix.lstrip('a').rstrip(':'))
					idx = agent_idxs[_idx]
					new_prefix = f'a{idx}'
					new_k = k.replace(prefix, new_prefix)
					eval_stats[new_k] = v
				else:
					eval_stats[k] = v

			for k,v in eval_stats.items():
				all_eval_stats[k].append(float(v))

			pbar.update(1)

	pbar.close()

	aggregate_eval_stats = {}
	for k,v in all_eval_stats.items():
		mean = np.mean(all_eval_stats[k])
		if len(all_eval_stats[k]) > 1:
			sem = spstats.sem(all_eval_stats[k])
		else:
			sem = 0.0
		aggregate_eval_stats[k] = f'{mean: 0.2}+/-{sem: 0.2}'

		_min = np.min(all_eval_stats[k])
		aggregate_eval_stats[f'min:{k}'] = f'{_min: 0.2}'

	logger = HumanOutputFormat(sys.stdout)
	logger.writekvs(aggregate_eval_stats)

	if args.results_fname is not None:
		if args.results_fname.strip('"') == '*':
			results_fname = args.xpid_prefix or args.xpid
		else:
			results_fname = args.results_fname

		df = pd.DataFrame.from_dict(all_eval_stats)
		results_path = args.results_path
		if not os.path.isabs(results_path):
			results_path = os.path.join(os.path.abspath(__file__), results_path)
		results_path = os.path.join(results_path, f'{results_fname}.csv')
		df.to_csv(results_path, index=False)
		print(f'Saved results to {results_path}')



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import numpy as np
import jax
import jax.numpy as jnp


class AgentPop:
	"""
	Manages multiple agents of the same architecture
	"""
	def __init__(
		self,
		agent,
		n_agents):
		"""
		Maintains a set of model parameters.
		"""
		self.agent = agent
		self.n_agents = n_agents

	def _reshape_to_pop(self, x):
		return jax.tree_map(
			lambda x: jnp.reshape(x, newshape=(self.n_agents, x.shape[0]//self.n_agents, *x.shape[1:])), x)

	def _flatten(self, x):
		return jax.tree_map(lambda x: jnp.reshape(x, newshape=(self.n_agents*x.shape[1], -1)).squeeze(), x)

	def init_params(self, rng, obs):
		if self.agent.is_recurrent:
			obs = jax.tree_map(lambda x: x[jnp.newaxis,:], obs) # Make time first dim

		vrngs = jax.random.split(rng, self.n_agents)
		return jax.vmap(
			self.agent.init_params, 
			in_axes=(0, None)
		)(vrngs, obs)

	@partial(jax.jit, static_argnums=(0,))
	def init_carry(self, rng, obs):
		agent_batch_dim = jax.tree_util.tree_leaves(obs)[0].shape[1]
		return self.agent.init_carry(rng=rng, batch_dims=(self.n_agents,agent_batch_dim))

	@partial(jax.jit, static_argnums=(0,))
	def act(self, params, obs, carry, reset=None):
		# If recurrent, add time axis to support scanned rollouts
		if self.agent.is_recurrent:
			obs = jax.tree_map(lambda x: x[:,jnp.newaxis,:], obs) # Add time dim after agent dim

			if reset is None:
				agent_batch_dim = jax.tree_util.tree_leaves(obs)[0].shape[2]
				reset = jnp.zeros((self.n_agents, 1, agent_batch_dim), dtype=jnp.bool_)
			else:
				reset = reset[:,jnp.newaxis,:]

		value, pi_params, next_carry = jax.vmap(self.agent.act)(params, obs, carry, reset)

		if self.agent.is_recurrent: # Remove time dim
			value = value.squeeze(1)
			pi_params =  jax.tree_map(lambda x: x.squeeze(1), pi_params) 

		return value, pi_params, next_carry

	def get_action_dist(self, dist_params, dtype=jnp.uint32):
		return self.agent.get_action_dist(dist_params, dtype=dtype)

	@partial(jax.jit, static_argnums=(0,))
	def get_value(self, params, obs, carry):
		return self.act(params, obs, carry)[0].squeeze(-1)

	@partial(jax.jit, static_argnums=(0,4,5))
	def update(self, rng, train_state, batch, prefix_steps=0, fake=False):
		if fake:
			return train_state, jax.vmap(lambda *_: self.agent.get_empty_update_stats())(np.arange(self.n_agents))

		rng, *vrngs = jax.random.split(rng, self.n_agents+1)
		vrngs = jnp.array(vrngs)
		
		return jax.vmap(self.agent.update)(vrngs, train_state, batch)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from typing import Any, Callable, Tuple
from collections import defaultdict, OrderedDict

import numpy as np
import jax
import jax.numpy as jnp
import flax
import optax
from flax.training.train_state import TrainState
from tensorflow_probability.substrates import jax as tfp

from .agent import Agent


class PPOAgent(Agent):
	def __init__(
		self,
		model,
		n_epochs=5,
		n_minibatches=1,
		value_loss_coef=0.5,
		entropy_coef=0.0,
		clip_eps=0.2,
		clip_value_loss=True,
		track_grad_norm=False,
		n_unroll_update=1,
		n_devices=1):

		self.model = model
		self.n_epochs = n_epochs
		self.n_minibatches = n_minibatches
		self.value_loss_coef = value_loss_coef
		self.entropy_coef = entropy_coef
		self.clip_eps = clip_eps
		self.clip_value_loss = clip_value_loss
		self.track_grad_norm = track_grad_norm
		self.n_unroll_update = n_unroll_update
		self.n_devices = n_devices

		self.grad_fn = jax.value_and_grad(self._loss, has_aux=True)

	@property
	def is_recurrent(self):
		return self.model.is_recurrent
		
	def init_params(self, rng, obs):
		"""
		Returns initialized parameters and RNN hidden state for a specific
		observation shape.
		"""
		rng, subrng = jax.random.split(rng)
		if self.model.is_recurrent:
			batch_size = jax.tree_util.tree_leaves(obs)[0].shape[1]
			carry = self.model.initialize_carry(rng=subrng, batch_dims=(batch_size,))
			reset = jnp.zeros((1, batch_size), dtype=jnp.bool_)

			rng, subrng = jax.random.split(rng)
			params = self.model.init(subrng, obs, carry, reset)
		else:
			params = self.model.init(subrng, obs)

		return params

	def init_carry(self, rng, batch_dims=(1,)):
		return self.model.initialize_carry(rng=rng, batch_dims=batch_dims)

	@partial(jax.jit, static_argnums=(0,))
	def act(self, params, obs, carry=None, reset=None):
		value, logits, carry = self.model.apply(params, obs, carry, reset) 

		return value, logits, carry

	@partial(jax.jit, static_argnums=(0,))
	def evaluate(self, params, action, obs, carry=None, reset=None):
		value, dist_params, carry = self.model.apply(params, obs, carry, reset)
		dist = self.get_action_dist(dist_params, dtype=action.dtype)
		log_prob = dist.log_prob(action)
		entropy = dist.entropy()

		return value.squeeze(), \
			log_prob.squeeze(), \
			entropy.squeeze(), \
			carry

	def get_action_dist(self, dist_params, dtype=jnp.uint32):
		return tfp.distributions.Categorical(logits=dist_params, dtype=dtype)

	@partial(jax.jit, static_argnums=(0,))
	def update(self, rng, train_state, batch):
		rngs = jax.random.split(rng, self.n_epochs)

		def _scan_epoch(carry, rng):
			brng, urng = jax.random.split(rng)
			batch, train_state = carry
			minibatches = self._get_minibatches(brng, batch)
			train_state, stats = \
				self._update_epoch(
					urng, train_state, minibatches)

			return (batch, train_state), stats

		(_, train_state), stats = jax.lax.scan(
			_scan_epoch,
			(batch, train_state),
			rngs,
			length=len(rngs)
		)

		stats = jax.tree_util.tree_map(lambda x: x.mean(), stats)
		train_state = train_state.increment_updates()

		return train_state, stats

	@partial(jax.jit, static_argnums=(0,))
	def get_empty_update_stats(self):
		keys = ['total_loss',
			'actor_loss',
			'value_loss',
			'entropy',
			'mean_value',
			'mean_target',
			'mean_gae',
			'grad_norm']

		return OrderedDict({k:-jnp.inf for k in keys})

	@partial(jax.jit, static_argnums=(0,))
	def _update_epoch(
		self,
		rng,
		train_state: TrainState,
		minibatches):

		def _update_minibatch(carry, step):
			rng, minibatch = step
			train_state = carry

			(loss, aux_info), grads = self.grad_fn(
				train_state.params,
				train_state.apply_fn,
				minibatch,
				rng,
			)

			loss_info = (loss,) + aux_info
			loss_info = loss_info + (optax.global_norm(grads),)

			if self.n_devices > 1:
				loss_info = jax.tree_map(lambda x: jax.lax.pmean(x, 'device'), loss_info)
				grads = jax.tree_map(lambda x: jax.lax.pmean(x, 'device'), grads)

			train_state = train_state.apply_gradients(grads=grads)

			stats_def = jax.tree_util.tree_structure(OrderedDict({
				k:0 for k in [
					'total_loss',
					'actor_loss',
					'value_loss',
					'entropy',
					'mean_value',
					'mean_target',
					'mean_gae',
					'grad_norm',
				]}))

			loss_stats = jax.tree_util.tree_unflatten(
				stats_def, jax.tree_util.tree_leaves(loss_info))

			return train_state, loss_stats

		rngs = jax.random.split(rng, self.n_minibatches)
		train_state, loss_stats = jax.lax.scan(
			_update_minibatch, 
			train_state, 
			(rngs, minibatches),
			length=self.n_minibatches,
			unroll=self.n_unroll_update
		)

		loss_stats = jax.tree_util.tree_map(lambda x: x.mean(axis=0), loss_stats)

		return train_state, loss_stats

	@partial(jax.jit, static_argnums=(0,2))
	def _loss(
		self, 
		params,
		apply_fn,
		batch,
		rng=None):
		carry = None

		if self.is_recurrent:
			"""
			Elements have batch shape of n_rollout_steps x n_envs//n_minibatches
			"""
			carry = jax.tree_util.tree_map(lambda x: x[0,:], batch.carry)
			obs, action, rewards, dones, log_pi_old, value_old, target, gae, carry_old = batch

			if self.is_recurrent:
				dones = dones.at[1:,:].set(dones[:-1,:])
				dones = dones.at[0,:].set(False)
				_batch = batch._replace(dones=dones)

				# Returns LxB and LxBxH tensors
				obs, action, _, done, _, _, _, _, _ = _batch
				value, log_pi, entropy, carry = apply_fn(params, action, obs, carry, done)
			else:
				value, log_pi, entropy, carry = apply_fn(params, action, obs, carry_old)
		else:
			obs, action, rewards, dones, log_pi_old, value_old, target, gae, _ = batch
			value, log_pi, entropy, _ = apply_fn(params, action, obs, carry)

		if self.clip_value_loss:
			value_pred_clipped = value_old + (value - value_old).clip(
				-self.clip_eps, self.clip_eps
			)
			value_losses = jnp.square(value - target)
			value_losses_clipped = jnp.square(value_pred_clipped - target)
			value_loss = 0.5 * jnp.maximum(value_losses, value_losses_clipped).mean()
		else:
			value_loss = optax.huber_loss(value, target).mean()

		if self.model.value_ensemble_size > 1:
			gae = gae.at[..., 0].get()

		ratio = jnp.exp(log_pi - log_pi_old)
		norm_gae = (gae - gae.mean()) / (gae.std() + 1e-5)
		loss_actor1 = ratio * norm_gae
		loss_actor2 = jnp.clip(ratio, 1.0 - self.clip_eps, 1.0 + self.clip_eps) * norm_gae
		loss_actor = -jnp.minimum(loss_actor1, loss_actor2).mean()

		entropy = entropy.mean()

		total_loss = (
			loss_actor + self.value_loss_coef*value_loss - self.entropy_coef*entropy
		)

		return total_loss, (
			loss_actor,
			value_loss,
			entropy,
			value.mean(),
			target.mean(),
			gae.mean()
		)

	@partial(jax.jit, static_argnums=0)
	def _get_minibatches(self, rng, batch):
		n_rollout_steps, n_envs = batch.dones.shape[0:2] # get dims based on dones
		if self.is_recurrent:
			"""
			Reshape elements into a batch shape of 
			n_minibatches x n_envs//n_minibatches x n_rollout_steps.
			"""
			assert n_envs % self.n_minibatches == 0, \
				'Number of environments must be divisible into number of minibatches.'

			n_env_per_minibatch = n_envs//self.n_minibatches
			shuffled_idx = jax.random.permutation(rng, jnp.arange(n_envs))

			shuffled_batch = jax.tree_util.tree_map(
				lambda x: jnp.take(x, shuffled_idx, axis=1), batch)

			minibatches = jax.tree_util.tree_map(
				lambda x: x.swapaxes(0,1).reshape(
					self.n_minibatches,
					n_env_per_minibatch,
					n_rollout_steps,
					*x.shape[2:]
				).swapaxes(1,2), shuffled_batch)
		else:
			n_txns = n_envs*n_rollout_steps
			assert n_envs*n_rollout_steps % self.n_minibatches == 0

			shuffled_idx = jax.random.permutation(rng, jnp.arange(n_txns))
			shuffled_batch = jax.tree_util.tree_map(
				lambda x: jnp.take(
					x.reshape(n_txns, *x.shape[2:]), 
						shuffled_idx, axis=0), batch)
			minibatches = jax.tree_util.tree_map(
				lambda x: x.reshape(self.n_minibatches,-1,*x.shape[1:]), shuffled_batch)

		return minibatches



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import pytest
import numpy as np
import jax
import jax.numpy as jnp

import util.rl.ued_scores as _ued_scores

from tests.base_req_rollout_storage import RequiresRolloutStorageTestClass


class TestUEDScores(RequiresRolloutStorageTestClass):
	def test_compute_ued_scores_returns(self):
		(obs, 
		 action, 
		 done, 
		 reward, 
		 log_pi, 
		 value, 
		 carry) = self._get_dummy_step_components()

		# Mark episode done every 8 steps
		batch_shape = self.batch_shape
		dones = jnp.zeros(batch_shape, dtype=jnp.uint8)
		dones = dones.at[:,jnp.arange(4,self.n_steps,4),:].set(1)

		# Reward of 10 at the end of every episode
		rewards = jnp.zeros_like(dones, dtype=jnp.float32)
		rewards = rewards.at[:,jnp.arange(4,self.n_steps,4),:].set(
			jnp.arange(1,8,dtype=jnp.float32).reshape(1,7,1)
		)

		rollout = self.rollout_mgr.reset()
		for t in range(self.n_steps):
			rollout = self.rollout_mgr.append(
				rollout,
				obs, 
				action, 
				rewards.at[:,t,:].get(),
				dones.at[:,t,:].get(),
				log_pi,
				value,
				carry
			)

		score_name = _ued_scores.UEDScore.RETURN
		last_value = jnp.zeros(self.t_batch_shape)
		batch = self.rollout_mgr.get_batch(rollout, last_value)
		ued_score, _ = _ued_scores.compute_ued_scores(
			score_name, batch, n_eval=self.n_eval)

		n_agents, n_steps, flat_batch_size = batch.dones.shape
		pop_batch_shape = (n_agents, n_steps, flat_batch_size//self.n_eval, self.n_eval)
		batch = jax.tree_util.tree_map(lambda x: x.reshape(*pop_batch_shape, *x.shape[3:]), batch)
		mean_env_returns_per_agent, max_env_returns_per_agent, _ = \
			jax.vmap(_ued_scores._compute_ued_scores, in_axes=(None, 0))(
				score_name, batch
			)

		mean_return = mean_env_returns_per_agent.mean(0)
		max_return = max_env_returns_per_agent.max(0)

		assert (mean_return != ued_score).sum() == 0

		batch_size = self.n_agents*self.n_envs
		assert mean_env_returns_per_agent.sum(0).sum(0) == jnp.arange(1,8).mean()*batch_size
		assert max_env_returns_per_agent.sum(0).sum(0) == 7*batch_size



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from .gridworld_models import (
	GridWorldACStudentModel, 
	GridWorldACTeacherModel,
)


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import jax.numpy as jnp

import minimax.envs as envs


class BatchEnv:
	def __init__(
		self,
		env_name,
		n_parallel,
		n_eval,
		env_kwargs,
		wrappers=['monitor_return']):
		self.env_name = env_name
		self.env, self.env_params = envs.make(
			env_name, 
			env_kwargs=env_kwargs,
			wrappers=wrappers,
		)
		self.n_parallel = n_parallel
		self.n_eval = n_eval

		self.sub_batch_size = n_parallel*n_eval

		self.step = jax.vmap(self._step, in_axes=0)
		self.get_env_metrics = jax.vmap(self._get_env_metrics, in_axes=0)
		self.set_state = jax.vmap(self._set_state, in_axes=0)

	@partial(jax.jit, static_argnums=(0,2,3))
	def reset(self, rng, n_parallel=None, n_eval=None):
		return jax.vmap(self._reset, in_axes=(0, None, None))(rng, n_parallel, n_eval)

	def _reset(self, rng, n_parallel=None, n_eval=None):
		# Create n_parallel envs, repeated n_eval times
		if n_parallel is None:
			n_parallel = self.n_parallel

		if n_eval is None:
			n_eval = self.n_eval

		brngs = jnp.repeat(jax.random.split(rng, n_parallel), n_eval, axis=0)

		obs, state, extra = jax.vmap(
			self.env.reset, in_axes=(0,))(brngs)

		return obs, state, extra

	@partial(jax.jit, static_argnums=0)
	def _step(self, rng, state, action, reset_state, extra):
		brngs = jax.random.split(rng, self.sub_batch_size)
		return jax.vmap(self.env.step, in_axes=(0, 0, 0, 0, 0))(
			brngs, state, action, reset_state, extra)

	@partial(jax.jit, static_argnums=(0,))
	def _get_env_metrics(self, state):
		return jax.vmap(self.env.get_env_metrics, in_axes=(0,))(state)

	@partial(jax.jit, static_argnums=(0,))
	def _set_state(self, state):
		# Need to repeat the state
		state = jax.tree_map(lambda x: x.repeat(self.n_eval, axis=0), state)

		return jax.vmap(self.env.set_state)(state)




"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from .env_wrapper import EnvWrapper
from .monitor_return import MonitorReturnWrapper
from .monitor_ep_metrics import MonitorEpisodicMetricsWrapper

from .ued_env_wrapper import UEDEnvWrapper


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from typing import Tuple, Optional

import jax
import jax.numpy as jnp
from flax import struct
import chex

from minimax.envs.registration import register
from .common import (
	DIR_TO_VEC,
	OBJECT_TO_INDEX,
	COLOR_TO_INDEX,
	make_maze_map,
)
from .maze import (
	Maze,
	EnvParams,
	EnvState,
	Actions
)

# ======== Singleton mazes ========
class MazeSingleton(Maze):
	def __init__(
		self,
		height=15,
		width=15,
		wall_map=None,
		goal_pos=None,
		agent_pos=None,
		agent_dir_idx=None,
		agent_view_size=5,
		see_through_walls=True,
		see_agent=False,
		normalize_obs=False,
		obs_agent_pos=False,
		max_episode_steps=None,
		singleton_seed=-1,
	):	
		super().__init__(
			height=height,
			width=width,
			agent_view_size=agent_view_size,
			see_through_walls=see_through_walls,
			see_agent=see_agent,
			normalize_obs=normalize_obs,
			obs_agent_pos=obs_agent_pos,
			max_episode_steps=max_episode_steps,
			singleton_seed=singleton_seed
		)

		if wall_map is None:
			self.wall_map = jnp.zeros((height,width), dtype=jnp.bool_)
		else:
			self.wall_map = \
				jnp.array(
					[[int(x) for x in row.split()] 
						for row in wall_map], dtype=jnp.bool_)
		height, width = self.wall_map.shape

		if max_episode_steps is None:
			max_episode_steps = 2*(height+2)*(width+2) # Match original eval steps

		self.goal_pos_choices = None
		if goal_pos is None:
			self.goal_pos = jnp.array([height, width]) - jnp.ones(2, dtype=jnp.uint32)
		elif isinstance(goal_pos, (tuple, list)) \
			and isinstance(goal_pos[0], (tuple, list)):
			self.goal_pos_choices = jnp.array(goal_pos, dtype=jnp.uint32)
			self.goal_pos = goal_pos[0]
		else:
			self.goal_pos = jnp.array(goal_pos, dtype=jnp.uint32)

		if agent_pos is None:
			self.agent_pos = jnp.zeros(2, dtype=jnp.uint32)
		else:
			self.agent_pos = jnp.array(agent_pos, dtype=jnp.uint32)

		self.agent_dir_idx = agent_dir_idx

		if self.agent_dir_idx is None:
			self.agent_dir_idx = 0

		self.params = EnvParams(
			height=height,
			width=width,
			agent_view_size=agent_view_size,
			see_through_walls=see_through_walls,
			see_agent=see_agent,
			normalize_obs=normalize_obs,
			obs_agent_pos=obs_agent_pos,
			max_episode_steps=max_episode_steps,
			singleton_seed=-1,
		)

		self.maze_map = make_maze_map( 
			self.params,
			self.wall_map, 
			self.goal_pos, 
			self.agent_pos, 
			self.agent_dir_idx,
			pad_obs=True)

	@property
	def default_params(self) -> EnvParams:
		# Default environment parameters
		return EnvParams()

	def reset_env(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.Array, EnvState]:

		if self.agent_dir_idx is None:
			key, subkey = jax.random.split(key)
			agent_dir_idx = jax.random.choice(subkey, 4)
		else:
			agent_dir_idx = self.agent_dir_idx

		if self.goal_pos_choices is not None:
			key, subkey = jax.random.split(key)
			goal_pos = jax.random.choice(subkey, self.goal_pos_choices)
			maze_map = make_maze_map( 
				self.params,
				self.wall_map, 
				goal_pos, 
				self.agent_pos, 
				agent_dir_idx,
				pad_obs=True)
		else:
			goal_pos = self.goal_pos
			maze_map = self.maze_map

		state = EnvState(
			agent_pos=self.agent_pos,
			agent_dir=DIR_TO_VEC[agent_dir_idx],
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=self.wall_map,
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state
		

# ======== Specific mazes ========
class SixteenRooms(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 1 0 0 1 0 0 1 0 0 0",
      "0 0 0 0 0 0 0 0 0 1 0 0 0",
      "0 0 0 1 0 0 1 0 0 0 0 0 0",
      "1 0 1 1 1 0 1 1 0 1 1 1 0",
      "0 0 0 1 0 0 0 0 0 0 0 0 0",
      "0 0 0 0 0 0 1 0 0 1 0 0 0",
      "1 1 0 1 0 1 1 0 1 1 1 0 1",
      "0 0 0 1 0 0 0 0 0 1 0 0 0",
      "0 0 0 1 0 0 1 0 0 0 0 0 0",
      "0 1 1 1 1 0 1 1 0 1 0 1 1",
      "0 0 0 1 0 0 1 0 0 1 0 0 0",
      "0 0 0 0 0 0 1 0 0 0 0 0 0",
      "0 0 0 1 0 0 0 0 0 1 0 0 0"
		]
		goal_pos = (11,11)
		agent_pos = (1,1)
		agent_dir_idx = 0

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class SixteenRooms2(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 1 0 0 0 0 0 1 0 0 0",
      "0 0 0 0 0 0 1 0 0 1 0 0 0",
      "0 0 0 1 0 0 1 0 0 1 0 0 0",
      "1 1 1 1 0 1 1 0 1 1 1 0 1",
      "0 0 0 1 0 0 1 0 0 0 0 0 0",
      "0 0 0 0 0 0 1 0 0 1 0 0 0",
      "1 0 1 1 1 1 1 0 1 1 1 1 1",
      "0 0 0 1 0 0 1 0 0 1 0 0 0",
      "0 0 0 1 0 0 0 0 0 0 0 0 0",
      "1 1 0 1 1 0 1 1 0 1 1 1 1",
      "0 0 0 1 0 0 1 0 0 1 0 0 0",
      "0 0 0 0 0 0 1 0 0 0 0 0 0",
      "0 0 0 1 0 0 1 0 0 1 0 0 0"
		]
		goal_pos = (11,11)
		agent_pos = (1,1)
		agent_dir_idx = None

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class Labyrinth(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 0 0 0 0 0 0 0 0 0 0",
      "0 1 1 1 1 1 1 1 1 1 1 1 0",
      "0 1 0 0 0 0 0 0 0 0 0 1 0",
      "0 1 0 1 1 1 1 1 1 1 0 1 0",
      "0 1 0 1 0 0 0 0 0 1 0 1 0",
      "0 1 0 1 0 1 1 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 0 0 1 0 0 0 1 0 1 0",
      "0 1 1 1 1 1 1 1 1 1 0 1 0",
      "0 0 0 0 0 1 0 0 0 0 0 1 0",
      "1 1 1 1 0 1 0 1 1 1 1 1 0",
      "0 0 0 0 0 1 0 0 0 0 0 0 0"
		]
		goal_pos = (6,6)
		agent_pos = (0,12)
		agent_dir_idx = 0

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class LabyrinthFlipped(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
			'0 0 0 0 0 0 0 0 0 0 0 0 0', 
			'0 1 1 1 1 1 1 1 1 1 1 1 0', 
			'0 1 0 0 0 0 0 0 0 0 0 1 0', 
			'0 1 0 1 1 1 1 1 1 1 0 1 0', 
			'0 1 0 1 0 0 0 0 0 1 0 1 0', 
			'0 1 0 1 0 1 1 1 0 1 0 1 0', 
			'0 1 0 1 0 1 0 1 0 1 0 1 0', 
			'0 1 0 1 0 1 0 1 0 1 0 1 0', 
			'0 1 0 1 0 0 0 1 0 0 0 1 0', 
			'0 1 0 1 1 1 1 1 1 1 1 1 0', 
			'0 1 0 0 0 0 0 1 0 0 0 0 0', 
			'0 1 1 1 1 1 0 1 0 1 1 1 1', 
			'0 0 0 0 0 0 0 1 0 0 0 0 0'
		]
		goal_pos = (6,6)
		agent_pos = (12,12)
		agent_dir_idx = 2

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)
		

class Labyrinth2(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 1 0 0 0 0 0 0 0 0 0 0 0",
      "0 1 0 1 1 1 1 1 1 1 1 1 0",
      "0 1 0 1 0 0 0 0 0 0 0 1 0",
      "0 1 0 1 0 1 1 1 1 1 0 1 0",
      "0 1 0 1 0 1 0 0 0 1 0 1 0",
      "0 0 0 1 0 1 0 1 0 1 0 1 0",
      "1 1 1 1 0 1 0 1 0 1 0 1 0",
      "0 0 0 1 0 1 1 1 0 1 0 1 0",
      "0 1 0 1 0 0 0 0 0 1 0 1 0",
      "0 1 0 1 1 1 1 1 1 1 0 1 0",
      "0 1 0 0 0 0 0 0 0 0 0 1 0",
      "0 1 1 1 1 1 1 1 1 1 1 1 0",
      "0 0 0 0 0 0 0 0 0 0 0 0 0"
		]
		goal_pos = (6,6)
		agent_pos = (0,0)
		agent_dir_idx = None

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class StandardMaze(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
			"0 0 0 0 0 1 0 0 0 0 1 0 0",
			"0 1 1 1 0 1 1 1 1 0 1 1 0",
			"0 1 0 0 0 0 0 0 0 0 0 0 0",
			"0 1 1 1 1 1 1 1 1 0 1 1 1",
			"0 0 0 0 0 0 0 0 1 0 0 0 0",
			"1 1 1 1 1 1 0 1 1 1 1 1 0",
			"0 0 0 0 1 0 0 1 0 0 0 0 0",
			"0 1 1 0 0 0 1 1 0 1 1 1 1",
			"0 0 1 0 1 0 0 1 0 0 0 1 0",
			"1 0 1 0 1 1 0 1 1 1 0 1 0",
			"1 0 1 0 0 1 0 0 0 1 0 0 0",
			"1 0 1 1 0 1 1 1 0 1 1 1 0",
			"0 0 0 1 0 0 0 1 0 1 0 0 0"
		]
		goal_pos = (6,12)
		agent_pos = (6,0)
		agent_dir_idx = 0

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class StandardMaze2(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 1 0 1 0 0 0 0 1 0 0",
      "0 1 0 1 0 1 1 1 1 0 0 0 1",
      "0 1 0 0 0 0 0 0 0 0 1 0 0",
      "0 1 1 1 1 1 1 1 1 0 1 1 1",
      "0 0 0 1 0 0 1 0 1 0 1 0 0",
      "1 1 0 1 0 1 1 0 1 0 1 0 0",
      "0 1 0 1 0 0 0 0 1 0 1 1 0",
      "0 1 0 1 1 0 1 1 1 0 0 1 0",
      "0 1 0 0 1 0 0 1 1 1 0 1 0",
      "0 1 1 0 1 1 0 1 0 1 0 1 0",
      "0 1 0 0 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 0 0 1 0 0 0 1 0 0 0 0 0"
		]
		goal_pos = (12,4)
		agent_pos = (0,6)
		agent_dir_idx = None

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class StandardMaze3(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 0 1 0 1 0 0 0 0 0 0",
      "0 1 1 1 1 0 1 0 1 1 1 1 0",
      "0 1 0 0 0 0 1 0 1 0 0 0 0",
      "0 0 0 1 1 1 1 0 1 0 1 0 1",
      "1 1 0 1 0 0 0 0 1 0 1 0 0",
      "0 0 0 1 0 1 1 0 1 0 1 1 0",
      "0 1 0 1 0 1 0 0 1 0 0 1 0",
      "0 1 0 1 0 1 0 1 1 1 0 1 1",
      "0 1 0 0 0 1 0 1 0 1 0 0 0",
      "0 1 1 1 0 1 0 1 0 1 1 1 0",
      "0 1 0 0 0 1 0 1 0 0 0 1 0",
      "0 1 0 1 1 1 0 1 0 1 0 1 0",
      "0 1 0 0 0 1 0 0 0 1 0 0 0"
		]
		goal_pos = (12,6)
		agent_pos = (3,0)
		agent_dir_idx = None

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class SmallCorridor(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 0 0 0 0 0 0 0 0 0 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 1 1 1 1 1 1 1 1 1 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 0 0 0 0 0 0 0 0 0 0 0 0"
		]
		goal_pos = [
			(2,5),(4,5),(6,5),(8,5),(10,5),
			(2,7),(4,7),(6,7),(8,7),(10,7),
		]
		agent_pos = (0,6)
		agent_dir_idx = None

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class LargeCorridor(MazeSingleton):
	def __init__(
		self, 
		see_agent=False,
		normalize_obs=False):
		wall_map = [
      "0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0",
      "0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0",
		]
		goal_pos = [
			(2,8),(4,8),(6,8),(8,8),(10,8),(12,8),(14,8),(16,8),
			(2,10),(4,10),(6,10),(8,10),(10,10),(12,10),(14,10),(16,10)
		]
		agent_pos = (0,9)
		agent_dir_idx = None

		super().__init__(
			wall_map=wall_map,
			goal_pos=goal_pos,
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			see_agent=see_agent,
			normalize_obs=normalize_obs
		)


class FourRooms(Maze):
	def __init__(
		self,
		height=17,
		width=17,
		agent_view_size=5,
		see_through_walls=True,
		see_agent=False,
		normalize_obs=False,
		max_episode_steps=250,
		singleton_seed=-1):

		super().__init__(
			height=height,
			width=width,
			agent_view_size=agent_view_size,
			see_through_walls=see_through_walls,
			see_agent=see_agent,
			normalize_obs=normalize_obs,
			max_episode_steps=max_episode_steps,
			singleton_seed=singleton_seed
		)

		assert height % 2 == 1 and width % 2 == 1, \
			'Grid height and width must be odd'

		wall_map = jnp.zeros((height, width), dtype=jnp.bool_)
		wall_map = wall_map.at[height//2, :].set(True)
		wall_map = wall_map.at[:, width//2].set(True)
		self.wall_map = wall_map

		self.room_h = height//2
		self.room_w = width//2

		self.all_pos_idxs = jnp.arange(height*width)
		self.goal_pos_mask = (~wall_map).flatten()
		self.agent_pos_mask = self.goal_pos_mask

	def reset_env(
		self, 
		key: chex.PRNGKey
	) -> Tuple[chex.Array, EnvState]:
		# Randomize door positions
		params = self.params

		key, x_rng, y_rng = jax.random.split(key,3)
		x_door_idxs = jax.random.randint(x_rng, (2,), 0, self.room_w) \
			+ jnp.array([0, self.room_w+1], dtype=jnp.uint32)

		y_door_idxs = jax.random.randint(y_rng, (2,), 0, self.room_h) \
			+ jnp.array([0, self.room_h+1], dtype=jnp.uint32)

		wall_map = self.wall_map.at[self.room_h, x_door_idxs].set(False)
		wall_map = wall_map.at[y_door_idxs,self.room_w].set(False)

		# Randomize goal pos
		key, subkey = jax.random.split(key)
		goal_pos_idx = jax.random.choice(subkey, self.all_pos_idxs, shape=(), p=self.goal_pos_mask)
		goal_pos = jnp.array([goal_pos_idx%params.width, goal_pos_idx//params.width], dtype=jnp.uint32)

		# Randomize agent pos
		key, subkey = jax.random.split(key)
		agent_pos_mask = self.agent_pos_mask.at[goal_pos_idx].set(False)
		agent_pos_idx = jax.random.choice(subkey, self.all_pos_idxs, shape=(), p=self.agent_pos_mask)
		agent_pos = jnp.array([agent_pos_idx%params.width, agent_pos_idx//params.width], dtype=jnp.uint32)

		key, subkey = jax.random.split(key)
		agent_dir_idx = jax.random.choice(subkey, 4)

		maze_map = make_maze_map( 
			self.params,
			wall_map, 
			goal_pos, 
			agent_pos, 
			agent_dir_idx,
			pad_obs=True)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=DIR_TO_VEC[agent_dir_idx],
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map,
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state


class Crossing(Maze):
	def __init__(
		self,
		height=9,
		width=9,
		n_crossings=5,
		agent_view_size=5,
		see_through_walls=True,
		see_agent=False,
		normalize_obs=False,
		max_episode_steps=250,
		singleton_seed=-1):
		self.n_crossings = n_crossings
		max_episode_steps = 4*(height+2)*(width+2)

		super().__init__(
			height=height,
			width=width,
			agent_view_size=agent_view_size,
			see_through_walls=see_through_walls,
			see_agent=see_agent,
			normalize_obs=normalize_obs,
			max_episode_steps=max_episode_steps,
			singleton_seed=singleton_seed
		)

	def reset_env(
		self, 
		key: chex.PRNGKey
	) -> Tuple[chex.Array, EnvState]:
		params = self.params
		height, width = params.height, params.width
		goal_pos = jnp.array([width-1, height-1])
		agent_pos = jnp.array([0,0], dtype=jnp.uint32)
		agent_dir_idx = 0

		# Generate walls
		wall_map = jnp.zeros((height, width), dtype=jnp.bool_)

		row_y_choices = jnp.arange(1,height-1,2)
		col_x_choices = jnp.arange(1,width-1,2)

		rng, subrng = jax.random.split(key)
		dirs = jax.random.permutation(
			subrng,
			jnp.concatenate(
				(jnp.zeros(len(row_y_choices)), 
				 jnp.ones(len(col_x_choices)))
			)
		)[:self.n_crossings]

		n_v = sum(dirs.astype(jnp.uint32))
		n_h = len(dirs) - n_v

		rng, row_rng, col_rng = jax.random.split(rng, 3)

		row_ys_mask = jax.random.permutation(row_rng, (jnp.arange(len(row_y_choices)) < n_v).repeat(2))
		if height % 2 == 0:
			row_ys_mask = jnp.concatenate((row_ys_mask, jnp.zeros(2)))
		else:
			row_ys_mask = jnp.concatenate((row_ys_mask, jnp.zeros(1)))

		row_ys_mask = jnp.logical_and(
			jnp.zeros(height, dtype=jnp.bool_).at[row_y_choices].set(True),
			row_ys_mask
		)

		col_xs_mask = jax.random.permutation(col_rng, (jnp.arange(len(col_x_choices)) < n_h).repeat(2))
		if width % 2 == 0:
			col_xs_mask = jnp.concatenate((col_xs_mask, jnp.zeros(2)))
		else:
			col_xs_mask = jnp.concatenate((col_xs_mask, jnp.zeros(1)))

		col_xs_mask = jnp.logical_and(
			jnp.zeros(width, dtype=jnp.bool_).at[col_x_choices].set(True),
			col_xs_mask
		)

		wall_map = jnp.logical_or(
			wall_map,
			jnp.tile(jnp.expand_dims(row_ys_mask,-1), (1,width))
		)

		wall_map = jnp.logical_or(
			wall_map,
			jnp.tile(jnp.expand_dims(col_xs_mask,0), (height,1))
		)

		# Generate wall openings
		def _scan_step(carry, rng):
			wall_map, pos, passed_wall, last_dir, last_dir_idx = carry

			dir_idx = jax.random.randint(rng,(),0,2)

			go_dir = (~passed_wall)*DIR_TO_VEC[dir_idx] + passed_wall*last_dir
			next_pos = pos + go_dir

			# If next pos is the right border, force direction to be down
			collide = jnp.logical_or(
				(next_pos[0] >= width),
				(next_pos[1] >= height)
			)
			go_dir = collide*DIR_TO_VEC[(dir_idx+1)%2] + (~collide)*go_dir
			dir_idx = (dir_idx+1)%2 + (~collide)*dir_idx

			next_pos = collide*(pos + go_dir) + (~collide)*next_pos

			last_dir = go_dir
			last_dir_idx = dir_idx
			pos = next_pos

			passed_wall = wall_map[pos[1],pos[0]]
			wall_map = wall_map.at[pos[1], pos[0]].set(False)

			return (wall_map, pos.astype(jnp.uint32), passed_wall, last_dir, last_dir_idx), None

		n_steps_to_goal = width + height - 2
		rng, *subrngs = jax.random.split(rng, n_steps_to_goal+1)

		pos = agent_pos
		passed_wall = jnp.array(False)
		last_dir = DIR_TO_VEC[0]

		(wall_map, pos, passed_wall, last_dir, last_dir_idx), _ = jax.lax.scan(
			_scan_step,
			(wall_map, pos, passed_wall, last_dir, 0),
			jnp.array(subrngs),
			length=n_steps_to_goal
		)

		maze_map = make_maze_map( 
			self.params,
			wall_map, 
			goal_pos, 
			agent_pos, 
			agent_dir_idx,
			pad_obs=True)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=DIR_TO_VEC[agent_dir_idx],
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map,
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state


NEIGHBOR_WALL_OFFSETS = jnp.array([
	[1,0], # right
	[0,1], # bottom
	[-1,0], # left
	[0,-1], # top
	[0,0] # self
], dtype=jnp.int32)


class PerfectMaze(Maze):
	def __init__(
		self,
		height=13,
		width=13,
		agent_view_size=5,
		see_through_walls=True,
		see_agent=False,
		normalize_obs=False,
		max_episode_steps=250,
		singleton_seed=-1):

		assert height % 2 == 1 and width % 2 == 1, \
			'Maze dimensions must be odd.'

		max_episode_steps = 2*(width+2)*(height+2)
		super().__init__(
			height=height,
			width=width,
			agent_view_size=agent_view_size,
			see_through_walls=see_through_walls,
			see_agent=see_agent,
			normalize_obs=normalize_obs,
			max_episode_steps=max_episode_steps,
			singleton_seed=singleton_seed
		)

	def reset_env(
		self, 
		key: chex.PRNGKey
	) -> Tuple[chex.Array, EnvState]:
		"""
		Generate a perfect maze using an iterated search procedure.
		"""
		params = self.params
		height, width = self.params.height, self.params.width
		n_tiles = height*width
		
		# Track maze wall map
		wall_map = jnp.ones((height, width), dtype=jnp.bool_)

		# Track visited, walkable tiles
		_h = height//2+1
		_w = width//2+1
		visited_map = jnp.zeros((_h, _w), dtype=jnp.bool_)
		vstack = jnp.zeros((_h*_w, 2), dtype=jnp.uint32)
		vstack_size = 0

		# Get initial start tile in walkable index
		key, subkey = jax.random.split(key)
		start_pos_x = jax.random.randint(subkey, (), 0, _w)
		start_pos_y = jax.random.randint(subkey, (), 0, _h)
		start_pos = jnp.array([start_pos_x,start_pos_y], dtype=jnp.uint32)

		# Set initial start tile as visited
		visited_map = visited_map.at[
			start_pos[1],start_pos[0]
		].set(True)
		wall_map = wall_map.at[
			2*start_pos[1],2*start_pos[0]
		].set(False)
		vstack = vstack.at[vstack_size:vstack_size+2].set(start_pos)
		vstack_size += 2

		def _scan_step(carry, key):
			# Choose last visited tile and move to a neighbor
			wall_map, visited_map, vstack, vstack_size = carry

			abs_pos = 2*vstack[vstack_size-1]

			neighbor_wall_offsets = NEIGHBOR_WALL_OFFSETS.at[-1].set(
				vstack[vstack_size-2] - vstack[vstack_size-1]
			)

			# Find a random unvisited neighbor
			neighbor_pos = \
				jnp.minimum(
					jnp.maximum(
						jnp.tile(abs_pos, (len(NEIGHBOR_WALL_OFFSETS),1)) \
						+ 2*neighbor_wall_offsets, 0
					), 
					jnp.array([width, height], dtype=jnp.uint32)
				)

			# Check for unvisited neighbors. Set self to unvisited if all visited.
			neighbor_visited = visited_map.at[
				neighbor_pos[:,1]//2, neighbor_pos[:,0]//2
			].get()

			n_neighbor_visited = neighbor_visited[:4].sum()
			all_visited = n_neighbor_visited == 4
			all_visited_post = n_neighbor_visited >= 3
			neighbor_visited = neighbor_visited.at[-1].set(~all_visited)

			# Choose a random unvisited neigbor and remove walls between current tile
			# and this neighbor and at this neighbor.
			rand_neighbor_idx = jax.random.choice(
				key, jnp.arange(len(NEIGHBOR_WALL_OFFSETS)), p=~neighbor_visited)
			rand_neighbor_pos = neighbor_pos[rand_neighbor_idx]
			rand_neighbor_wall_pos = abs_pos + (~all_visited)*neighbor_wall_offsets[rand_neighbor_idx]
			remove_wall_pos = jnp.concatenate(
				(jnp.expand_dims(rand_neighbor_pos, 0), 
				 jnp.expand_dims(rand_neighbor_wall_pos,0)), 0)
			wall_map = wall_map.at[
				remove_wall_pos[:,1], remove_wall_pos[:,0]
			].set(False)

			# Set selected neighbor as visited
			visited_map = visited_map.at[
				rand_neighbor_pos[1]//2,rand_neighbor_pos[0]//2
			].set(True)

			# Pop current tile from stack if all neighbors have been visited
			vstack_size -= all_visited_post

			# Push selected neighbor onto stack
			vstack = vstack.at[vstack_size].set(
				rand_neighbor_pos//2
			)
			vstack_size += ~all_visited

			return (wall_map, visited_map, vstack, vstack_size), None

		# for i in range(3*_h*_w):
		max_n_steps = 2*_w*_h
		key, *subkeys = jax.random.split(key, max_n_steps+1)
		(wall_map, visited_map, vstack, vstack_size), _ = jax.lax.scan(
			_scan_step,
			(wall_map, visited_map, vstack, vstack_size),
			jnp.array(subkeys),
			length=max_n_steps
		)

		# Randomize goal position
		all_pos_idx = jnp.arange(height*width)

		key, subkey = jax.random.split(key)
		goal_mask = ~wall_map.flatten()
		goal_pos_idx = jax.random.choice(subkey, all_pos_idx, p=goal_mask)
		goal_pos = jnp.array([goal_pos_idx%width, goal_pos_idx//width])

		# Randomize agent position
		key, subkey = jax.random.split(key)
		agent_mask = goal_mask.at[goal_pos_idx].set(False)
		agent_pos_idx = jax.random.choice(subkey, all_pos_idx, p=agent_mask)
		agent_pos = jnp.array([agent_pos_idx%width, agent_pos_idx//width], dtype=jnp.uint32)

		# Randomize agent dir
		key, subkey = jax.random.split(key)
		agent_dir_idx = jax.random.choice(subkey, 4)

		maze_map = make_maze_map( 
			self.params,
			wall_map, 
			goal_pos, 
			agent_pos, 
			agent_dir_idx,
			pad_obs=True)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=DIR_TO_VEC[agent_dir_idx],
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map,
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state


class PerfectMazeMedium(PerfectMaze):
	def __init__(self, *args, **kwargs):
		super().__init__(height=19, width=19, *args, **kwargs)


class PerfectMazeExtraLarge(PerfectMaze):
	def __init__(self, *args, **kwargs):
		super().__init__(height=101, width=101, *args, **kwargs)


class Memory(MazeSingleton):
	def __init__(
		self,
		height=17,
		width=17,
		agent_view_size=7,
		see_through_walls=True,
		see_agent=False,
		normalize_obs=False,
		obs_agent_pos=False,
		max_episode_steps=250,
		singleton_seed=-1):

		# Generate walls
		wall_map = [
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "1 1 1 1 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 1 1 1 1 1 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 0 0 1 0 0 0 0",
	      "0 0 0 1 1 1 1 1 1 0 1 0 0 0 0",
	      "1 1 1 1 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0",
	      "0 0 0 0 0 0 0 0 1 0 1 0 0 0 0"
		]

		super().__init__(
			wall_map=wall_map,
			goal_pos=(9,5),
			agent_pos=(0,7),
			agent_dir_idx=0,
			see_agent=see_agent,
			normalize_obs=normalize_obs,
			obs_agent_pos=obs_agent_pos,
			max_episode_steps=max_episode_steps
		)

		self.top_pos = jnp.array([9,5], dtype=jnp.uint32)
		self.bottom_pos = jnp.array([9,9], dtype=jnp.uint32)

	def reset_env(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.Array, EnvState]:
		params = self.params
		height, width = params.height, params.width

		agent_pos = jnp.array([0,7], dtype=jnp.uint32)
		agent_dir_idx = 0

		# Randomly generate a memory location
		is_top_goal = jax.random.randint(key, minval=0, maxval=2, shape=(1,), dtype=jnp.uint8)

		clue_pos = jnp.array((0,6), dtype=jnp.uint32)
		self.goal_pos = is_top_goal*self.top_pos + (1-is_top_goal)*self.bottom_pos
		self.distractor_pos = is_top_goal*self.bottom_pos + (1-is_top_goal)*self.top_pos

		goal_color = is_top_goal*COLOR_TO_INDEX['red'] + (1-is_top_goal)*COLOR_TO_INDEX['green']

		wall_map = self.wall_map
		maze_map = make_maze_map( 
			self.params,
			jnp.array(wall_map, dtype=jnp.bool_), 
			self.goal_pos, 
			agent_pos, 
			agent_dir_idx,
			pad_obs=True)

		red_goal = jnp.array([OBJECT_TO_INDEX['goal'], COLOR_TO_INDEX['red'], 0], dtype=jnp.uint8)
		green_goal = jnp.array([OBJECT_TO_INDEX['goal'], COLOR_TO_INDEX['green'], 0], dtype=jnp.uint8)
		clue = is_top_goal*red_goal + (1-is_top_goal)*green_goal

		padding = params.agent_view_size-1
		wall_map = wall_map.at[clue_pos[1], clue_pos[0]].set(True)
		maze_map = maze_map.at[padding+clue_pos[1], padding+clue_pos[0]].set(clue)

		wall_map = wall_map.at[self.top_pos[1], self.top_pos[0]].set(True)
		maze_map = maze_map.at[padding+self.top_pos[1], padding+self.top_pos[0]].set(red_goal)

		wall_map = wall_map.at[self.bottom_pos[1], self.bottom_pos[0]].set(True)
		maze_map = maze_map.at[padding+self.bottom_pos[1], padding+self.bottom_pos[0]].set(green_goal)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=DIR_TO_VEC[agent_dir_idx],
			agent_dir_idx=agent_dir_idx,
			goal_pos=self.goal_pos,
			wall_map=wall_map,
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state

	def get_distractor_pos(self, state):
		goal_x, goal_y = state.goal_pos
		is_top_goal = jnp.logical_and(
			goal_x == self.top_pos[0],
			goal_y == self.top_pos[1]
		)

		return is_top_goal*self.bottom_pos + (1-is_top_goal)*self.top_pos


	def step_agent(self, key: chex.PRNGKey, state: EnvState, action: int) -> Tuple[EnvState, float]:
		next_state, reward = super().step_agent(
			key=key,
			state=state,
			action=action
		)

		fwd_pos = jnp.minimum(
			jnp.maximum(state.agent_pos + (action == Actions.forward)*state.agent_dir, 0), 
			jnp.array((
				self.params.width-1, self.params.height-1), 
				dtype=jnp.uint32
			))

		distractor_pos = self.get_distractor_pos(state)
		fwd_pos_has_distractor = jnp.logical_and(
			fwd_pos[0] == distractor_pos[0], 
			fwd_pos[1] == distractor_pos[1]
		)

		next_state = next_state.replace(
			terminal=jnp.logical_or(
				next_state.terminal, 
				fwd_pos_has_distractor
			)
		)

		return (
			next_state,
			reward
		)


# ======== Registration ========
if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register(env_id='Maze-SixteenRooms', entry_point=module_path + ':SixteenRooms')
register(env_id='Maze-SixteenRooms2', entry_point=module_path + ':SixteenRooms2')
register(env_id='Maze-Labyrinth', entry_point=module_path + ':Labyrinth')
register(env_id='Maze-Labyrinth2', entry_point=module_path + ':Labyrinth2')
register(env_id='Maze-LabyrinthFlipped', entry_point=module_path + ':LabyrinthFlipped')
register(env_id='Maze-StandardMaze', entry_point=module_path + ':StandardMaze')
register(env_id='Maze-StandardMaze2', entry_point=module_path + ':StandardMaze2')
register(env_id='Maze-StandardMaze3', entry_point=module_path + ':StandardMaze3')
register(env_id='Maze-SmallCorridor', entry_point=module_path + ':SmallCorridor')
register(env_id='Maze-LargeCorridor', entry_point=module_path + ':LargeCorridor')

register(env_id='Maze-FourRooms', entry_point=module_path + ':FourRooms')
register(env_id='Maze-Crossing', entry_point=module_path + ':Crossing')
register(env_id='Maze-PerfectMaze', entry_point=module_path + ':PerfectMaze')
register(env_id='Maze-PerfectMazeMedium', entry_point=module_path + ':PerfectMazeMedium')
register(env_id='Maze-PerfectMazeXL', entry_point=module_path + ':PerfectMazeExtraLarge')

register(env_id='Maze-Memory', entry_point=module_path + ':Memory')


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This file is modified from
https://github.com/luchris429/purejaxrl

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0
"""

from functools import partial
import jax
import jax.numpy as np
import jax.numpy as jnp
from flax import linen as nn
from jax.nn.initializers import lecun_normal, normal
from jax import random
from jax.numpy.linalg import eigh
from jax.scipy.linalg import block_diag


def log_step_initializer(dt_min=0.001, dt_max=0.1):
	""" Initialize the learnable timescale Delta by sampling
		 uniformly between dt_min and dt_max.
		 Args:
			 dt_min (float32): minimum value
			 dt_max (float32): maximum value
		 Returns:
			 init function
	 """
	def init(key, shape):
		""" Init function
			 Args:
				 key: jax random key
				 shape tuple: desired shape
			 Returns:
				 sampled log_step (float32)
		 """
		return random.uniform(key, shape) * (
			np.log(dt_max) - np.log(dt_min)
		) + np.log(dt_min)

	return init


def init_log_steps(key, input):
	""" Initialize an array of learnable timescale parameters
		 Args:
			 key: jax random key
			 input: tuple containing the array shape H and
					dt_min and dt_max
		 Returns:
			 initialized array of timescales (float32): (H,)
	 """
	H, dt_min, dt_max = input
	log_steps = []
	for i in range(H):
		key, skey = random.split(key)
		log_step = log_step_initializer(dt_min=dt_min, dt_max=dt_max)(skey, shape=(1,))
		log_steps.append(log_step)

	return np.array(log_steps)


def init_VinvB(init_fun, rng, shape, Vinv):
	""" Initialize B_tilde=V^{-1}B. First samples B. Then compute V^{-1}B.
		Note we will parameterize this with two different matrices for complex
		numbers.
		 Args:
			 init_fun:  the initialization function to use, e.g. lecun_normal()
			 rng:       jax random key to be used with init function.
			 shape (tuple): desired shape  (P,H)
			 Vinv: (complex64)     the inverse eigenvectors used for initialization
		 Returns:
			 B_tilde (complex64) of shape (P,H,2)
	 """
	B = init_fun(rng, shape)
	VinvB = Vinv @ B
	VinvB_real = VinvB.real
	VinvB_imag = VinvB.imag
	return np.concatenate((VinvB_real[..., None], VinvB_imag[..., None]), axis=-1)


def trunc_standard_normal(key, shape):
	""" Sample C with a truncated normal distribution with standard deviation 1.
		 Args:
			 key: jax random key
			 shape (tuple): desired shape, of length 3, (H,P,_)
		 Returns:
			 sampled C matrix (float32) of shape (H,P,2) (for complex parameterization)
	 """
	H, P, _ = shape
	Cs = []
	for i in range(H):
		key, skey = random.split(key)
		C = lecun_normal()(skey, shape=(1, P, 2))
		Cs.append(C)
	return np.array(Cs)[:, 0]


def init_CV(init_fun, rng, shape, V):
	""" Initialize C_tilde=CV. First sample C. Then compute CV.
		Note we will parameterize this with two different matrices for complex
		numbers.
		 Args:
			 init_fun:  the initialization function to use, e.g. lecun_normal()
			 rng:       jax random key to be used with init function.
			 shape (tuple): desired shape  (H,P)
			 V: (complex64)     the eigenvectors used for initialization
		 Returns:
			 C_tilde (complex64) of shape (H,P,2)
	 """
	C_ = init_fun(rng, shape)
	C = C_[..., 0] + 1j * C_[..., 1]
	CV = C @ V
	CV_real = CV.real
	CV_imag = CV.imag
	return np.concatenate((CV_real[..., None], CV_imag[..., None]), axis=-1)


# Discretization functions
def discretize_bilinear(Lambda, B_tilde, Delta):
	""" Discretize a diagonalized, continuous-time linear SSM
		using bilinear transform method.
		Args:
			Lambda (complex64): diagonal state matrix              (P,)
			B_tilde (complex64): input matrix                      (P, H)
			Delta (float32): discretization step sizes             (P,)
		Returns:
			discretized Lambda_bar (complex64), B_bar (complex64)  (P,), (P,H)
	"""
	Identity = np.ones(Lambda.shape[0])

	BL = 1 / (Identity - (Delta / 2.0) * Lambda)
	Lambda_bar = BL * (Identity + (Delta / 2.0) * Lambda)
	B_bar = (BL * Delta)[..., None] * B_tilde
	return Lambda_bar, B_bar


def discretize_zoh(Lambda, B_tilde, Delta):
	""" Discretize a diagonalized, continuous-time linear SSM
		using zero-order hold method.
		Args:
			Lambda (complex64): diagonal state matrix              (P,)
			B_tilde (complex64): input matrix                      (P, H)
			Delta (float32): discretization step sizes             (P,)
		Returns:
			discretized Lambda_bar (complex64), B_bar (complex64)  (P,), (P,H)
	"""
	Identity = np.ones(Lambda.shape[0])
	Lambda_bar = np.exp(Lambda * Delta)
	B_bar = (1/Lambda * (Lambda_bar-Identity))[..., None] * B_tilde
	return Lambda_bar, B_bar


# Parallel scan operations
@jax.vmap
def binary_operator(q_i, q_j):
	""" Binary operator for parallel scan of linear recurrence. Assumes a diagonal matrix A.
		Args:
			q_i: tuple containing A_i and Bu_i at position i       (P,), (P,)
			q_j: tuple containing A_j and Bu_j at position j       (P,), (P,)
		Returns:
			new element ( A_out, Bu_out )
	"""
	A_i, b_i = q_i
	A_j, b_j = q_j
	return A_j * A_i, A_j * b_i + b_j

# Parallel scan operations
@jax.vmap
def binary_operator_reset(q_i, q_j):
	""" Binary operator for parallel scan of linear recurrence. Assumes a diagonal matrix A.
		Args:
			q_i: tuple containing A_i and Bu_i at position i       (P,), (P,)
			q_j: tuple containing A_j and Bu_j at position j       (P,), (P,)
		Returns:
			new element ( A_out, Bu_out )
	"""
	A_i, b_i, c_i = q_i
	A_j, b_j, c_j = q_j
	return (
		(A_j * A_i)*(1 - c_j) + A_j * c_j,
		(A_j * b_i + b_j)*(1 - c_j) + b_j * c_j,
		c_i * (1 - c_j) + c_j,
	)


def apply_ssm(Lambda_bar, B_bar, C_tilde, hidden, input_sequence, resets, conj_sym, bidirectional):
	""" Compute the LxH output of discretized SSM given an LxH input.
		Args:
			Lambda_bar (complex64): discretized diagonal state matrix    (P,)
			B_bar      (complex64): discretized input matrix             (P, H)
			C_tilde    (complex64): output matrix                        (H, P)
			input_sequence (float32): input sequence of features         (L, H)
			reset      (bool): input sequence of features                (L,)
			conj_sym (bool):         whether conjugate symmetry is enforced
			bidirectional (bool):    whether bidirectional setup is used,
								  Note for this case C_tilde will have 2P cols
		Returns:
			ys (float32): the SSM outputs (S5 layer preactivations)      (L, H)
	"""
	Lambda_elements = Lambda_bar * jnp.ones((input_sequence.shape[0],
											Lambda_bar.shape[0]))
	Bu_elements = jax.vmap(lambda u: B_bar @ u)(input_sequence)

	Lambda_elements = jnp.concatenate([
		jnp.ones((1, Lambda_bar.shape[0])),
		Lambda_elements,
	])

	Bu_elements = jnp.concatenate([
		hidden,
		Bu_elements,
	])

	resets = jnp.concatenate([
		jnp.zeros(1),
		resets,
	])

	_, xs, _ = jax.lax.associative_scan(binary_operator_reset, (Lambda_elements, Bu_elements, resets))
	xs = xs[1:]

	if conj_sym:
		return xs[np.newaxis, -1], jax.vmap(lambda x: 2*(C_tilde @ x).real)(xs)
	else:
		return xs[np.newaxis, -1], jax.vmap(lambda x: (C_tilde @ x).real)(xs)


class S5SSM(nn.Module):
	Lambda_re_init: jax.Array
	Lambda_im_init: jax.Array
	V: jax.Array
	Vinv: jax.Array

	H: int
	P: int
	C_init: str
	discretization: str
	dt_min: float
	dt_max: float
	conj_sym: bool = True
	clip_eigs: bool = False
	bidirectional: bool = False
	step_rescale: float = 1.0

	""" The S5 SSM
		Args:
			Lambda_re_init (complex64): Real part of init diag state matrix  (P,)
			Lambda_im_init (complex64): Imag part of init diag state matrix  (P,)
			V           (complex64): Eigenvectors used for init           (P,P)
			Vinv        (complex64): Inverse eigenvectors used for init   (P,P)
			H           (int32):     Number of features of input seq 
			P           (int32):     state size
			C_init      (string):    Specifies How C is initialized
						 Options: [trunc_standard_normal: sample from truncated standard normal 
														and then multiply by V, i.e. C_tilde=CV.
								   lecun_normal: sample from Lecun_normal and then multiply by V.
								   complex_normal: directly sample a complex valued output matrix 
													from standard normal, does not multiply by V]
			conj_sym    (bool):    Whether conjugate symmetry is enforced
			clip_eigs   (bool):    Whether to enforce left-half plane condition, i.e.
								   constrain real part of eigenvalues to be negative. 
								   True recommended for autoregressive task/unbounded sequence lengths
								   Discussed in https://arxiv.org/pdf/2206.11893.pdf.
			bidirectional (bool):  Whether model is bidirectional, if True, uses two C matrices
			discretization: (string) Specifies discretization method 
							 options: [zoh: zero-order hold method,
									   bilinear: bilinear transform]
			dt_min:      (float32): minimum value to draw timescale values from when 
									initializing log_step
			dt_max:      (float32): maximum value to draw timescale values from when 
									initializing log_step
			step_rescale:  (float32): allows for uniformly changing the timescale parameter, e.g. after training 
									on a different resolution for the speech commands benchmark
	"""

	def setup(self):
		"""Initializes parameters once and performs discretization each time
		   the SSM is applied to a sequence
		"""
		if self.conj_sym:
			# Need to account for case where we actually sample real B and C, and then multiply
			# by the half sized Vinv and possibly V
			local_P = 2*self.P
		else:
			local_P = self.P

		# Initialize diagonal state to state matrix Lambda (eigenvalues)
		self.Lambda_re = self.param("Lambda_re", lambda rng, shape: self.Lambda_re_init, (None,))
		self.Lambda_im = self.param("Lambda_im", lambda rng, shape: self.Lambda_im_init, (None,))
		if self.clip_eigs:
			self.Lambda = np.clip(self.Lambda_re, None, -1e-4) + 1j * self.Lambda_im
		else:
			self.Lambda = self.Lambda_re + 1j * self.Lambda_im

		# Initialize input to state (B) matrix
		B_init = lecun_normal()
		B_shape = (local_P, self.H)
		self.B = self.param("B",
							lambda rng, shape: init_VinvB(B_init,
														  rng,
														  shape,
														  self.Vinv),
							B_shape)
		B_tilde = self.B[..., 0] + 1j * self.B[..., 1]

		# Initialize state to output (C) matrix
		if self.C_init in ["trunc_standard_normal"]:
			C_init = trunc_standard_normal
			C_shape = (self.H, local_P, 2)
		elif self.C_init in ["lecun_normal"]:
			C_init = lecun_normal()
			C_shape = (self.H, local_P, 2)
		elif self.C_init in ["complex_normal"]:
			C_init = normal(stddev=0.5 ** 0.5)
		else:
			raise NotImplementedError(
				   "C_init method {} not implemented".format(self.C_init))

		if self.C_init in ["complex_normal"]:
			if self.bidirectional:
				C = self.param("C", C_init, (self.H, 2 * self.P, 2))
				self.C_tilde = C[..., 0] + 1j * C[..., 1]

			else:
				C = self.param("C", C_init, (self.H, self.P, 2))
				self.C_tilde = C[..., 0] + 1j * C[..., 1]

		else:
			if self.bidirectional:
				self.C1 = self.param("C1",
									 lambda rng, shape: init_CV(C_init, rng, shape, self.V),
									 C_shape)
				self.C2 = self.param("C2",
									 lambda rng, shape: init_CV(C_init, rng, shape, self.V),
									 C_shape)

				C1 = self.C1[..., 0] + 1j * self.C1[..., 1]
				C2 = self.C2[..., 0] + 1j * self.C2[..., 1]
				self.C_tilde = np.concatenate((C1, C2), axis=-1)

			else:
				self.C = self.param("C",
									lambda rng, shape: init_CV(C_init, rng, shape, self.V),
									C_shape)

				self.C_tilde = self.C[..., 0] + 1j * self.C[..., 1]

		# Initialize feedthrough (D) matrix
		self.D = self.param("D", normal(stddev=1.0), (self.H,))

		# Initialize learnable discretization timescale value
		self.log_step = self.param("log_step",
								   init_log_steps,
								   (self.P, self.dt_min, self.dt_max))
		step = self.step_rescale * np.exp(self.log_step[:, 0])

		# Discretize
		if self.discretization in ["zoh"]:
			self.Lambda_bar, self.B_bar = discretize_zoh(self.Lambda, B_tilde, step)
		elif self.discretization in ["bilinear"]:
			self.Lambda_bar, self.B_bar = discretize_bilinear(self.Lambda, B_tilde, step)
		else:
			raise NotImplementedError("Discretization method {} not implemented".format(self.discretization))

	def __call__(self, hidden, input_sequence, resets):
		"""
		Compute the LxH output of the S5 SSM given an LxH input sequence
		using a parallel scan.
		Args:
			 input_sequence (float32): input sequence (L, H)
			 resets (bool): input sequence (L,)
		Returns:
			output sequence (float32): (L, H)
		"""
		hidden, ys = apply_ssm(
			self.Lambda_bar,
			self.B_bar,
			self.C_tilde,
			hidden,
			input_sequence,
			resets,
			self.conj_sym,
			self.bidirectional)
		# Add feedthrough matrix output Du;
		Du = jax.vmap(lambda u: self.D * u)(input_sequence)
		return hidden, ys + Du


class SequenceLayer(nn.Module):
	""" Defines a single S5 layer, with S5 SSM, nonlinearity,
			dropout, batch/layer norm, etc.
		Args:
			ssm         (nn.Module): the SSM to be used (i.e. S5 ssm)
			dropout     (float32):  dropout rate
			d_model     (int32):    this is the feature size of the layer inputs and outputs
									we usually refer to this size as H
			activation  (string):   Type of activation function to use
			training    (bool):     whether in training mode or not
			prenorm     (bool):     apply prenorm if true or postnorm if false
			batchnorm   (bool):     apply batchnorm if true or layernorm if false
			bn_momentum (float32):  the batchnorm momentum if batchnorm is used
			step_rescale  (float32):  allows for uniformly changing the timescale parameter,
									e.g. after training on a different resolution for
									the speech commands benchmark
	"""
	ssm: nn.Module
	d_model: int
	activation: str = "gelu"
	layernorm_pos: str = None # ['pre', 'post', None]
	step_rescale: float = 1.0

	def setup(self):
		"""Initializes the ssm, batch/layer norm and dropout
		"""
		self.seq = self.ssm(step_rescale=self.step_rescale)

		if self.activation in ["full_glu"]:
			self.out1 = nn.Dense(self.d_model)
			self.out2 = nn.Dense(self.d_model)
		elif self.activation in ["half_glu1", "half_glu2"]:
			self.out2 = nn.Dense(self.d_model)

		self.norm = nn.LayerNorm()
		self.drop = lambda x: x

	def __call__(self, hidden, x, d):
		"""
		Compute the LxH output of S5 layer given an LxH input.
		Args:
			 x (float32): input sequence (L, B, d_model)
			 d (bool): reset signal (L,B)
		Returns:
			output sequence (float32): (L, B, d_model)
		"""
		is_one_step = len(hidden.shape) == 2
		if is_one_step: # Add time axis
			hidden = hidden[jnp.newaxis,:]

		skip = x
		if self.layernorm_pos == 'pre':
		    x = self.norm(x)
		hidden, x = jax.vmap(self.seq, in_axes=1, out_axes=1)(hidden, x, d)

		if self.activation in ["full_glu"]:
			x = self.drop(nn.gelu(x))
			x = self.out1(x) * jax.nn.sigmoid(self.out2(x))
			x = self.drop(x)
		elif self.activation in ["half_glu1"]:
			x = self.drop(nn.gelu(x))
			x = x * jax.nn.sigmoid(self.out2(x))
			x = self.drop(x)
		elif self.activation in ["half_glu2"]:
			# Only apply GELU to the gate input
			x1 = self.drop(nn.gelu(x))
			x = x * jax.nn.sigmoid(self.out2(x1))
			x = self.drop(x)
		elif self.activation in ["gelu"]:
			x = self.drop(nn.gelu(x))
		else:
			raise NotImplementedError(
				   "Activation: {} not implemented".format(self.activation))

		x = skip + x
		if self.layernorm_pos == 'post':
		    x = self.norm(x)
		if is_one_step:
			hidden = hidden.squeeze(0)

		return hidden, x

	@staticmethod
	def initialize_carry(batch_size, hidden_size):
		return jnp.zeros((batch_size, hidden_size), dtype=jnp.complex64)


def init_S5SSM(
	H,
	P,
	Lambda_re_init,
	Lambda_im_init,
	V,
	Vinv,
	C_init,
	discretization,
	dt_min,
	dt_max,
	conj_sym,
	clip_eigs,
	bidirectional):
	"""Convenience function that will be used to initialize the SSM.
	   Same arguments as defined in S5SSM above."""
	return partial(S5SSM,
		H=H,
		P=P,
		Lambda_re_init=Lambda_re_init,
		Lambda_im_init=Lambda_im_init,
		V=V,
		Vinv=Vinv,
		C_init=C_init,
		discretization=discretization,
		dt_min=dt_min,
		dt_max=dt_max,
		conj_sym=conj_sym,
		clip_eigs=clip_eigs,
		bidirectional=bidirectional)


def make_HiPPO(N):
	""" Create a HiPPO-LegS matrix.
		From https://github.com/srush/annotated-s4/blob/main/s4/s4.py
		Args:
			N (int32): state size
		Returns:
			N x N HiPPO LegS matrix
	"""
	P = np.sqrt(1 + 2 * np.arange(N))
	A = P[:, np.newaxis] * P[np.newaxis, :]
	A = np.tril(A) - np.diag(np.arange(N))
	return -A


def make_NPLR_HiPPO(N):
	"""
	Makes components needed for NPLR representation of HiPPO-LegS
	 From https://github.com/srush/annotated-s4/blob/main/s4/s4.py
	Args:
		N (int32): state size
	Returns:
		N x N HiPPO LegS matrix, low-rank factor P, HiPPO input matrix B
	"""
	# Make -HiPPO
	hippo = make_HiPPO(N)

	# Add in a rank 1 term. Makes it Normal.
	P = np.sqrt(np.arange(N) + 0.5)

	# HiPPO also specifies the B matrix
	B = np.sqrt(2 * np.arange(N) + 1.0)
	return hippo, P, B


def make_DPLR_HiPPO(N):
	"""
	Makes components needed for DPLR representation of HiPPO-LegS
	 From https://github.com/srush/annotated-s4/blob/main/s4/s4.py
	Note, we will only use the diagonal part
	Args:
		N:
	Returns:
		eigenvalues Lambda, low-rank term P, conjugated HiPPO input matrix B,
		eigenvectors V, HiPPO B pre-conjugation
	"""
	A, P, B = make_NPLR_HiPPO(N)

	S = A + P[:, np.newaxis] * P[np.newaxis, :]

	S_diag = np.diagonal(S)
	Lambda_real = np.mean(S_diag) * np.ones_like(S_diag)

	# Diagonalize S to V \Lambda V^*
	Lambda_imag, V = eigh(S * -1j)

	P = V.conj().T @ P
	B_orig = B
	B = V.conj().T @ B
	return Lambda_real + 1j * Lambda_imag, P, B, V, B_orig

class S5EncoderStack(nn.Module):
	""" Defines a stack of S5 layers to be used as an encoder.
		Args:
			ssm         (nn.Module): the SSM to be used (i.e. S5 ssm)
			d_model     (int32):    this is the feature size of the layer inputs and outputs
									 we usually refer to this size as H
			n_layers    (int32):    the number of S5 layers to stack
			activation  (string):   Type of activation function to use
			dropout     (float32):  dropout rate
			training    (bool):     whether in training mode or not
			prenorm     (bool):     apply prenorm if true or postnorm if false
			batchnorm   (bool):     apply batchnorm if true or layernorm if false
			bn_momentum (float32):  the batchnorm momentum if batchnorm is used
			step_rescale  (float32):  allows for uniformly changing the timescale parameter,
									e.g. after training on a different resolution for
									the speech commands benchmark
	"""
	ssm: nn.Module
	d_model: int
	n_layers: int
	layernorm_pos: str = None
	activation: str = "gelu"

	def setup(self):
		"""
		Initializes a linear encoder and the stack of S5 layers.
		"""
		self.layers = [
			SequenceLayer(
				ssm=self.ssm,
				d_model=self.d_model,
				activation=self.activation,
				layernorm_pos=self.layernorm_pos,
			)
			for _ in range(self.n_layers)
		]

	def __call__(self, hidden, x, reset):
		"""
		Compute the BxLxH output of the stacked encoder given an Lxd_input
		input sequence.
		Args:
			 x (float32): input sequence (L, d_input)
		Returns:
			output sequence (float32): (L, d_model)
		"""
		new_hiddens = []
		for i, layer in enumerate(self.layers):
			new_h, x = layer(hidden[i], x, reset)
			new_hiddens.append(new_h)
	
		return new_hiddens, x

	@staticmethod
	def initialize_carry(rng, batch_dims, hidden_dim, n_layers):
		# Use a dummy key since the default state init fn is just zeros.
		return [jnp.zeros((*batch_dims, hidden_dim), dtype=jnp.complex64) for _ in range(n_layers)]


BatchS5EncoderStack = nn.vmap(
	S5EncoderStack,
	in_axes=(1, 1, 1),
	out_axes=1,
	variable_axes={"params": None},
	split_rngs={"params": False}, axis_name='batch')


def make_s5_encoder_stack(
	input_dim, 
	ssm_state_dim,
	n_blocks=1, 
	n_layers=4,
	discretization='zoh', 
	dt_min=0.001, 
	dt_max=0.1, 
	conj_sym=True, 
	clip_eigs=False, 
	bidirectional=False,
	activation="half_glu1",
	layernorm_pos=None):
	block_size = int(ssm_state_dim / n_blocks)

	Lambda, _, B, V, B_orig = make_DPLR_HiPPO(block_size)

	if conj_sym:
		block_size = block_size // 2
		ssm_state_dim = ssm_state_dim // 2

	Lambda = Lambda[:block_size]
	V = V[:, :block_size]
	Vc = V.conj().T

	Lambda = (Lambda*np.ones((n_blocks, block_size))).ravel()
	V = block_diag(*([V]*n_blocks))
	Vinv = block_diag(*([Vc]*n_blocks))

	ssm_init_fn = init_S5SSM(
		H=input_dim,
		P=ssm_state_dim,
		Lambda_re_init=Lambda.real,
		Lambda_im_init=Lambda.imag,
		V=V,
		Vinv=Vinv,
		C_init="lecun_normal",
		discretization=discretization,
		dt_min=dt_min,
		dt_max=dt_max,
		conj_sym=conj_sym,
		clip_eigs=clip_eigs,
		bidirectional=bidirectional)

	return S5EncoderStack(
		ssm=ssm_init_fn,
		d_model=input_dim,
		n_layers=n_layers,
		activation=activation,
		layernorm_pos=layernorm_pos
	)
